// QuickHeap.h
#pragma once

#define QUICK_HEAP_SYNCHRONIZATION

#ifdef QUICK_HEAP_SYNCHRONIZATION
#include <atlcore.h>
struct CQHAutoLock
{
	CComAutoCriticalSection & m_cs;
	CQHAutoLock(CComAutoCriticalSection & cs) : m_cs(cs){ cs.Lock(); }
	~CQHAutoLock(){ m_cs.Unlock(); }
};
#endif // QUICK_HEAP_SYNCHRONIZATION

//         
//   ASC_MEM_STATISTIC_USE_QUICK_HEAP   
// OnQuickHeapInternalAlloc  OnQuickHeapPoolInternalFree
//  #include "QuickHeap.h"
inline void * QuickHeapInternalAlloc(size_t size)
{
	void * p = HeapAlloc(GetProcessHeap(), HEAP_NO_SERIALIZE, size);
	//FillMemory(p, size, 0xfe);
#ifdef ASC_MEM_STATISTIC_USE_QUICK_HEAP
	OnQuickHeapInternalAlloc(p, size);
#endif	//ASC_MEM_STATISTIC_USE_QUICK_HEAP
	return p;
}
inline void * QuickHeapInternalReAlloc(void * p, size_t size)
{
	void * pRet = HeapReAlloc(GetProcessHeap(), HEAP_NO_SERIALIZE, p, size);
#ifdef ASC_MEM_STATISTIC_USE_QUICK_HEAP
	ATLASSERT(0);
#endif	//ASC_MEM_STATISTIC_USE_QUICK_HEAP
	return pRet;
}
inline void QuickHeapPoolInternalFree(void * pBlock)
{
#ifdef ASC_MEM_STATISTIC_USE_QUICK_HEAP
	OnQuickHeapPoolInternalFree(pBlock);
#endif	//ASC_MEM_STATISTIC_USE_QUICK_HEAP
	HeapFree(GetProcessHeap(), HEAP_NO_SERIALIZE, pBlock);
}


struct QHBlock
{
	union
	{
		QHBlock * m_pNext;			//   
		void * m_pQuickHeapPool;	//   ( )
	};
protected:
	inline void * __cdecl operator new(size_t n){ n; }
	inline void   __cdecl operator delete(void* p){ p; }
};

struct QHExtent
{
	QHExtent * m_pNextExtent;
	// QHBlock * m_pCurrent; -   QuickHeapPool  m_pCurrentBlock;
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4200)
#endif
	QHBlock m_aryBlocks[0];
#ifdef _MSC_VER
#pragma warning(pop)
#endif
protected:
	inline void * __cdecl operator new(size_t n){ n; }
	inline void   __cdecl operator delete(void* p){ p; }
};

class QuickHeapPool
{
public:
	inline void * __cdecl operator new(size_t n)
	{ return QuickHeapInternalAlloc(n); }
	inline void   __cdecl operator delete(void* p)
	{ QuickHeapPoolInternalFree(p); }
	void * Alloc()
	{
		QHBlock * p = AllocBlock();
		p++;
		return p;
	}
	void Free(void * p)
	{
		ATLASSERT(0);
		FreeBlock((QHBlock*)(BYTE(p) - sizeof(QHBlock*)));
	}
	QuickHeapPool(size_t BlockSize, int iInitBlockCnt = 0) 
		: m_CurrExtent(NULL),m_pCurrentBlock(NULL),m_pFreeBlocksList(NULL)
	{
		m_iIntrBlockSize = sizeof(QHBlock) + BlockSize;
		if(iInitBlockCnt > 0)
			AddExtent(iInitBlockCnt);
	}
	~QuickHeapPool(void)
	{
		QHExtent * pExtent = m_CurrExtent;
		while(pExtent)
		{
			QHExtent * pNext = pExtent->m_pNextExtent;
			QuickHeapPoolInternalFree(pExtent);
			pExtent = pNext;
		}
	}
	__forceinline QHBlock * AllocBlock()
	{
		if(m_pFreeBlocksList)
		{
			register QHBlock * pBlock = m_pFreeBlocksList;
			m_pFreeBlocksList = m_pFreeBlocksList->m_pNext;
			pBlock->m_pQuickHeapPool = this;
			return pBlock;
		}
		else
			return GetBlock();
	}
	__forceinline void FreeBlock(QHBlock * pBlock)
	{
		pBlock->m_pNext = m_pFreeBlocksList;
		m_pFreeBlocksList = pBlock;
	}
	__forceinline size_t GetBlockSize()
	{	//    m_iIntrBlockSize - sizeof(QHBlock)
		return m_iIntrBlockSize;
	}
protected:
	void AddExtent(size_t iBlockCnt)
	{
		ATLASSERT(!m_CurrExtent 
			|| m_pCurrentBlock == m_CurrExtent->m_aryBlocks);
		QHExtent * pExtent = (QHExtent*)QuickHeapInternalAlloc(
			sizeof(QHExtent) + (iBlockCnt/* + 1*/) * m_iIntrBlockSize);
		pExtent->m_pNextExtent = m_CurrExtent;
		ATLASSERT(iBlockCnt > 0);

		// - 1 - ..       .
		//   -1        .
		m_pCurrentBlock = (QHBlock*)((BYTE*)(pExtent->m_aryBlocks) 
			+ (iBlockCnt - 1) * m_iIntrBlockSize); 

		m_CurrExtent = pExtent;
		m_iCurrExtentSize = iBlockCnt;
	}
	__forceinline QHBlock * GetBlock()
	{
		if(m_pCurrentBlock == m_CurrExtent->m_aryBlocks)
			AddExtent(m_iCurrExtentSize * 2);
		QHBlock * pRet = m_pCurrentBlock;
		pRet->m_pQuickHeapPool = this;
		(BYTE*&)m_pCurrentBlock -= m_iIntrBlockSize;
		return pRet;
	}
	//  .
	size_t m_iCurrExtentSize;
	size_t m_iIntrBlockSize;
	QHExtent * m_CurrExtent;
	QHBlock * m_pCurrentBlock;
	QHBlock * m_pFreeBlocksList;
};

const QUICK_HEAP_REALLOC_COPY_DATA = FALSE;
const QUICK_HEAP_REALLOC_FILL_ZERO = FALSE;
const ciQHInitPoolSizeDefault = 0x1000;	//    pool'e
const ciQHInitPoolArraySizeDefault = 1024; //!  ,  ciQHInitPoolSizeDefault!
class QuickHeap
{
public:
	inline void * __cdecl operator new(size_t n)
	{ return QuickHeapInternalAlloc(n); }
	inline void __cdecl operator delete(void* p)
	{ QuickHeapPoolInternalFree(p); }
	QuickHeap()
	{
		PoolSizeLoad();
		const iCountByte = m_iQHInitPoolArraySize * sizeof(QuickHeapPool *);
		m_arypQHPool = (QuickHeapPool **) QuickHeapInternalAlloc(iCountByte);
		ZeroMemory(m_arypQHPool, iCountByte);
	}
	~QuickHeap()
	{
		for(int i = 0; i < m_iQHInitPoolArraySize; i++)
			delete (m_arypQHPool[i]);
		PoolSizeSave();
	}
	__forceinline void * Alloc(size_t cb)
	{
#ifdef QUICK_HEAP_SYNCHRONIZATION
		CQHAutoLock qhal(m_cs);
#endif // QUICK_HEAP_SYNCHRONIZATION
		const int iMaxHeapSize = m_iQHInitPoolArraySize;
		if(cb >= (size_t)iMaxHeapSize)
		{
			QHBlock * pBlock = (QHBlock*)QuickHeapInternalAlloc(
				sizeof(QHBlock) + cb);
			pBlock->m_pQuickHeapPool = (QuickHeapPool*)-1;
			++pBlock;
			return pBlock;

		}
		else if(0 == cb) // cb    < 0, .. size_t  DWORD
			return NULL;
		else
		{
			QuickHeapPool * pCurrPool = m_arypQHPool[cb];
			if(!pCurrPool)
				pCurrPool = m_arypQHPool[cb] = 
					QuickHeapPoolNew(cb, m_iQHInitPoolSize / (int)cb);
			return pCurrPool->Alloc();
		}
	}
	__forceinline void Free(void * p)
	{
#ifdef QUICK_HEAP_SYNCHRONIZATION
		CQHAutoLock qhal(m_cs);
#endif // QUICK_HEAP_SYNCHRONIZATION
		QHBlock * pBlock = (QHBlock *)p;
		pBlock--;
		QuickHeapPool* pQuickHeapPool = (QuickHeapPool*)pBlock->m_pQuickHeapPool;
		if((QuickHeapPool*)-1 == pQuickHeapPool)
			QuickHeapPoolInternalFree(pBlock);
		//  ,     .
		//else if(NULL == pQuickHeapPool)
		//	return;
		else
			pQuickHeapPool->FreeBlock(pBlock);
	}
	__forceinline void * ReAlloc(void * p, size_t size,
		BOOL bCopyData = QUICK_HEAP_REALLOC_COPY_DATA,
		BOOL bFillZero = QUICK_HEAP_REALLOC_FILL_ZERO
	)
	{
#ifdef QUICK_HEAP_SYNCHRONIZATION
		CQHAutoLock qhal(m_cs);
#endif // QUICK_HEAP_SYNCHRONIZATION
		void * pNew = Alloc(size);
		ATLASSERT(pNew);
		if(p)
		{
#ifdef	_DEBUG
			// !   HeapValidate!
			//		if(!HeapValidate(GetProcessHeap(), HEAP_NO_SERIALIZE, p))
			//		{	//    (,   heap'a)
			//			ATLASSERT(!"Invalid heap memory allocation address!");
			//		}
#endif	_DEBUG
			if(bCopyData)
			{	//      
				QHBlock * pBlock = (QHBlock *)p;
				pBlock--;
				size_t dwSizeOld = ((QuickHeapPool*)(pBlock->m_pQuickHeapPool))->GetBlockSize();
				if(dwSizeOld == (DWORD)-1)
					return QuickHeapInternalReAlloc(p, size);
				dwSizeOld -= sizeof(QHBlock);
				if(dwSizeOld)
				{
					CopyMemory(pNew, pBlock, dwSizeOld);
					if(bFillZero && size > dwSizeOld)
					{	//   ,     
						//    ,      
						ZeroMemory((void*)((int)pNew + dwSizeOld), size - dwSizeOld);
					}
				}
			}
			Free(p);
		}
		return pNew;
	}
	//       
	// !    HeapSize  p!
	//__forceinline size_t GetBlockSize(void * p)
	//{
	//	QHBlock * pBlock = (QHBlock *)p;
	//	pBlock--;
	//	QuickHeapPool* pQuickHeapPool = (QuickHeapPool*)pBlock->m_pQuickHeapPool;
	//	return pQuickHeapPool->GetBlockSize() - sizeof(QHBlock);
	//}
protected:
	//    , ..   new  
	//  __forceinline   Alloc.
	//   ,     inline
	inline QuickHeapPool * QuickHeapPoolNew(size_t BlockSize, int iInitBlockCnt)
	{
		return new QuickHeapPool(BlockSize, iInitBlockCnt);
	}
#define REG_VALUE_QHINITPOOLSIZE _T("QHInitPoolSize")
#define REG_VALUE_QHINITPOOLARRAYSIZE _T("QHInitPoolArraySize")
	inline HKEY RegKeyCreate()
	{
		DWORD dw = NULL;
		HKEY hKey = NULL;
		LONG lr = RegCreateKeyEx(
			HKEY_LOCAL_MACHINE, _T("SOFTWARE\\Optim.ru\\QuickHeap"), 0,
			REG_NONE, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 
			(LPSECURITY_ATTRIBUTES)NULL, &hKey, &dw);
		HRESULT hr = HRESULT_FROM_WIN32(lr);
		ATLASSERT(SUCCEEDED(hr));hr;
		return hKey;
	}
	inline void RegKeyClose(HKEY hKey)
	{
		LONG lr = RegCloseKey(hKey);
		HRESULT hr = HRESULT_FROM_WIN32(lr);
		ATLASSERT(!hr);hr;
	}
	inline void RegKeyValuePoolSizeGet(HKEY hKey)
	{
			DWORD dwValue = 0;
			DWORD dwType = 0;
			DWORD dwCount = sizeof(DWORD);
			LONG lr = RegQueryValueEx(
				hKey, (LPTSTR)REG_VALUE_QHINITPOOLSIZE, NULL,
				&dwType, (LPBYTE)&dwValue, &dwCount);
			HRESULT hr = HRESULT_FROM_WIN32(lr);
			BOOL bOk = SUCCEEDED(hr) && (dwType == REG_DWORD) && (dwCount == sizeof(DWORD));
			if(bOk)
				m_iQHInitPoolSize = (int)dwValue;
			else
				m_iQHInitPoolSize = ciQHInitPoolSizeDefault;

	}
	inline void RegKeyValuePoolArraySizeGet(HKEY hKey)
	{
			DWORD dwValue = 0;
			DWORD dwType = 0;
			DWORD dwCount = sizeof(DWORD);
			LONG lr = RegQueryValueEx(
				hKey, (LPTSTR)REG_VALUE_QHINITPOOLARRAYSIZE, NULL,
				&dwType, (LPBYTE)&dwValue, &dwCount);
			HRESULT hr = HRESULT_FROM_WIN32(lr);
			BOOL bOk = SUCCEEDED(hr) && (dwType == REG_DWORD) && (dwCount == sizeof(DWORD));
			if(bOk)
				m_iQHInitPoolArraySize = (int)dwValue;
			else
				m_iQHInitPoolArraySize = ciQHInitPoolArraySizeDefault;
	}
	inline void RegKeyValuePoolSizeSet(HKEY hKey)
	{
			int iValue = m_iQHInitPoolSize;
			LONG lr = RegSetValueEx(
				hKey, (LPTSTR)REG_VALUE_QHINITPOOLSIZE, NULL, REG_DWORD,
				(BYTE * const)&iValue, sizeof(DWORD));
			HRESULT hr = HRESULT_FROM_WIN32(lr);
			ATLASSERT(SUCCEEDED(hr)); hr;
	}
	inline void RegKeyValuePoolArraySizeSet(HKEY hKey)
	{
			int iValue = m_iQHInitPoolArraySize;
			LONG lr = RegSetValueEx(
				hKey, (LPTSTR)REG_VALUE_QHINITPOOLARRAYSIZE, NULL, REG_DWORD,
				(BYTE * const)&iValue, sizeof(DWORD));
			HRESULT hr = HRESULT_FROM_WIN32(lr);
			ATLASSERT(SUCCEEDED(hr)); hr;
	}
	//    ,  ,       pool'
	inline void PoolSizeLoad()
	{
		HKEY hKey = RegKeyCreate();
		if(hKey)
		{
			RegKeyValuePoolSizeGet(hKey);
			RegKeyValuePoolArraySizeGet(hKey);
			RegKeyClose(hKey);
		}
		else
		{
			m_iQHInitPoolSize = ciQHInitPoolSizeDefault;
			m_iQHInitPoolArraySize = ciQHInitPoolArraySizeDefault;
		}
		//     ,  -   pool'e
		ATLASSERT(m_iQHInitPoolSize > m_iQHInitPoolArraySize && m_iQHInitPoolArraySize > 0 && m_iQHInitPoolArraySize > 0);
		if(m_iQHInitPoolArraySize >= m_iQHInitPoolSize)
		{
			if(1 == m_iQHInitPoolSize)
				++m_iQHInitPoolSize;
			m_iQHInitPoolArraySize = m_iQHInitPoolSize - 1;
		}
	}
	inline void PoolSizeSave()
	{
		HKEY hKey = RegKeyCreate();
		if(hKey)
		{
			RegKeyValuePoolSizeSet(hKey);
			RegKeyValuePoolArraySizeSet(hKey);
			RegKeyClose(hKey);
		}
	}
public:
	//QuickHeapPool * m_arypQHPool[ciQHInitPoolArraySizeDefault];
	QuickHeapPool ** m_arypQHPool; //  pool'     1  N
#ifdef QUICK_HEAP_SYNCHRONIZATION
	CComAutoCriticalSection m_cs;
#endif // QUICK_HEAP_SYNCHRONIZATION
protected:
	int m_iQHInitPoolSize;
	int m_iQHInitPoolArraySize;
};



/*
#ifndef DEBUG
extern "C"
{
	void _endthread( void )
	{
		MessageBox(NULL, "_endthread", NULL, 0);
	}
	void _endthreadex( unsigned retval )
	{
		retval;
		MessageBox(NULL, "_endthreadex", NULL, 0);
	}
	unsigned long _beginthread( void( __cdecl *start_address )( void * ), unsigned stack_size, void *arglist )
	{
		start_address, stack_size, arglist;
		MessageBox(NULL, "_beginthread", NULL, 0);
		return 0;
	}

	unsigned long _beginthreadex( void *security, unsigned stack_size, unsigned ( __stdcall *start_address )( void * ), void *arglist, unsigned initflag, unsigned *thrdaddr )
	{
		security, stack_size, start_address, arglist, initflag, thrdaddr;
		MessageBox(NULL, "_beginthreadex", NULL, 0);
		return 0;
	}
}
#endif //DEBUG
//*/

