/*
 * Purpose: Implement Java version of algorithms participate in the performance testing
 *
 * Created on August 23, 2001, 9:59 PM
 */

import java.io.*;
import java.util.Arrays;

class CBase
{
    public CBase(int i)
    {
        m_i = i;
        m_pGreat = null;
        m_pLess = null;
        m_pEqual = null;
    }
    final int m_i;
    CBase m_pGreat;
    CBase m_pLess;
    CBase m_pEqual;
    private final static void AddNode(CBase pRoot, CBase pNew)
    {
        if(pNew.m_i > pRoot.m_i)
        {
            if(pRoot.m_pGreat != null)
                AddNode(pRoot.m_pGreat, pNew);
            else
                pRoot.m_pGreat = pNew;
        }
        else if(pNew.m_i < pRoot.m_i)
        {
            if(pRoot.m_pLess != null)
                AddNode(pRoot.m_pLess, pNew);
            else
                pRoot.m_pLess = pNew;
        }
        else
        {
            if(pRoot.m_pEqual != null)
                AddNode(pRoot.m_pEqual, pNew);
            else
                pRoot.m_pEqual = pNew;
        }
    }

    public final static void TreeSort(int[] item, int left, int right)
    {
        CTestEqualZero Root = new CTestEqualZero(0);
        CBase pNewNode = null;
        for(int i = left; i <= right; i++)
        {
            int iCurrItem = item[i];
            if(iCurrItem < 0)
                pNewNode = new CTestLessZero(iCurrItem);
            else if(iCurrItem == 0)
                pNewNode = new CTestEqualZero(iCurrItem);
            else
                pNewNode = new CTestGreatZero(iCurrItem);
            AddNode(Root, pNewNode);
        }
    }
};

class CTestGreatZero extends CBase
{
    public CTestGreatZero(int i)
    {
      super(i);
    }
    protected final int m_i1 = 0;
    protected final int m_i2 = 0;
    protected final int m_i3 = 0;
};

class CTestLessZero extends CTestGreatZero
{
    public CTestLessZero(int i)
    {
      super(i);
    }
    public final int m_i4 = 0;
    public final int m_i5 = 0;
};

class CTestEqualZero extends CTestLessZero
{
    public CTestEqualZero(int i)
    {
      super(i);
    }
    private final int m_i6 = 0;
    private final double m_d = 0;
};


/**
 * @author  Alexey Zorkaltsev (a.k.a. ZORK)
 * @version 1.0
 */
public class PerformanceTest {

    /** Creates new PerformanceTest */
    public PerformanceTest() {
    }

    static final int c_iCount = 1000000000; // Number of iterations in member and method access tests

    static final String DATA_FILE_NAME = "..\\q3.zip";

    public static void main(String args[]) throws Exception
    {
/*	  System.out.println("Loading data ...");
	  int[] iAryData = loadData(DATA_FILE_NAME, 100000000);

	System.out.println("Start Quick sort test!");
	System.out.println("Quick sort took " + sortTest(iAryData, ST_QUICK_SORT, 100000000, 1) / 1000.0 + " seconds");

	System.out.println("Start Bubble sort test!");
	System.out.println("Bubble sort took " + sortTest(iAryData, ST_BUBBLE_SORT, 150000, 1) / 1000.0 + " seconds");

	System.out.println("Start Tree sort test!");
	System.out.println("Tree sort took " + sortTest(iAryData, ST_TREE_SORT, 10000000, 1) / 1000.0 + " seconds");
	System.out.println("Start Tree sort test 2!");
	System.out.println("Tree sort took " + sortTest(iAryData, ST_TREE_SORT, 10000000, 1) / 1000.0 + " seconds");
	System.out.println("Start Tree sort test 3!");
	System.out.println("Tree sort took " + sortTest(iAryData, ST_TREE_SORT, 10000000, 1) / 1000.0 + " seconds");
	System.out.println("Start Tree sort test 4!");
	System.out.println("Tree sort took " + sortTest(iAryData, ST_TREE_SORT, 10000000, 1) / 1000.0 + " seconds");

	System.out.println("reset data...");
	iAryData = null;

	System.out.println("Start Pi test...");
	System.out.println("Pi computation test took " + piComputationTest(15000) / 1000.0 + " seconds");

	System.out.println("Start Member Access test...");
	System.out.println("Member Access test took " + memberAccessTest() / 1000.0 + " seconds");

	System.out.println("Start Method Access test...");
	System.out.println("Method Access test took " + methodCallTest() / 1000.0 + " seconds");

	System.out.println("Virtual Method Access test took " + VirtualMethodCallTest() / 1000.0 + " seconds");

	System.out.println("String Test took " + StringTest() / 1000.0 + " seconds");

	System.out.println("Float Test took " + FloatTest() / 1000.0 + " seconds");
*/
	System.out.println("Str. Atrr took " + StrAtrTest() / 1000.0 + " seconds");
    }

    ////////////////////////////////////////////////////////
    // Member Access and Method Call Tests

    static class CTest1 // ZORK: 'final' and 'static' might affect performance!
    {
        int m_i = 0;
        public int GetInt() {
            return m_i;
        }
        public final void Inc() {
            m_i++;
        }
        public final void Test(int i, double d, short sh, String sz)
        {
            //i, d, sh, sz;
        }
        public void TestVirtual(int i, double d, short sh, String sz)
        {
            //i, d, sh, sz;
        }
    };

    static final long memberAccessTest() {
        CTest1 test1 = new CTest1();
        long start = System.currentTimeMillis();
        for(int i = 0; i < c_iCount; i++) {
            test1.Inc();
        }
        long stop = System.currentTimeMillis();
        return stop - start;
    }

    static final void methodCall(CTest1 test1, int iInitVal)
    {
        for(int i = 0; i < iInitVal; i++)
            test1.Test(i, 3.123, (short)1, "Some string");
    }

	//    final ,   
    static void VirtualMethodCall(CTest1 test1, int iInitVal)
    {
        for(int i = 0; i < iInitVal; i++)
            test1.TestVirtual(i, 3.123, (short)1, "Some string");
    }

    static final long methodCallTest()
    {
        CTest1 test1 = new CTest1();
        long start = System.currentTimeMillis();
        methodCall(test1, c_iCount);
        long stop = System.currentTimeMillis();
        return stop - start;
    }

    static final long VirtualMethodCallTest()
    {
        CTest1 test1 = new CTest1();
        long start = System.currentTimeMillis();
        VirtualMethodCall(test1, c_iCount);
        long stop = System.currentTimeMillis();
        return stop - start;
    }
    ////////////////////////////////////////////////////////
    // Sort Test, including Quick and Bubble sort

    static final int ST_QUICK_SORT = 1;
    static final int ST_BUBBLE_SORT = 2;
    static final int ST_TREE_SORT = 3;

    static int[] loadData(java.lang.String sFileName, int iFileSize) throws IOException
    {
        final int ciElemCnt = iFileSize / 4 /* sizeof(int) */; // ZORK: I guess say 'final' is a reasonable optimization
	System.out.println("Allocating " + ciElemCnt + " elems of int32...");
        int[] iAryData = new int[ciElemCnt];
        {
	    System.out.println("Array allocated! Loading data.");
            DataInput di;
            di = new java.io.DataInputStream(new BufferedInputStream(new FileInputStream(sFileName)));
            for (int i = 0; i < ciElemCnt; ++i)
                iAryData[i] = di.readInt();
	    System.out.println("Data loaded!");
        }
        return iAryData;
    }

    static long sortTest(int[] iAryData, int eSortType, int ciByts, int iIterationsCnt)
    {

        final int ciElemCnt = ciByts / 4; // sizeof(int32)
        int[] iAryWork;

        iAryWork = (int[])iAryData.clone();
        long start = System.currentTimeMillis();
        switch (eSortType) {
            case ST_QUICK_SORT:
                for(int i = 0; i < iIterationsCnt; i++) {
                    if(i > 0)
                        iAryWork = (int[])iAryData.clone();
                    quickSort(iAryWork, 0, ciElemCnt - 1);
                }
                break;
            case ST_BUBBLE_SORT:
                for(int i = 0; i < iIterationsCnt; i++) {
                    if(i > 0)
                        iAryWork = (int[])iAryData.clone();
                    bubbleSort(iAryWork, 0, ciElemCnt - 1);
                }
                break;
            case ST_TREE_SORT:
                for(int i = 0; i < iIterationsCnt; i++) {
                    if(i > 0)
                        iAryWork = (int[])iAryData.clone();
                    CBase.TreeSort(iAryWork, 0, ciElemCnt - 1);
                }
                break;
        }
        long stop = System.currentTimeMillis();
        return stop - start;
    }

    static final void quickSort(int[] item, int left, int right) {
        int i, j;
        int center;
        int x;
        i = left;
        j = right;
        center = item[(left + right) / 2];
        while(i <= j) {
            while (item[i] < center && i < right)
                i++;
            while (item[j] > center && j > left)
                j--;
            if (i<=j) {
                x  = item[i];
                item[i] = item[j];
                item[j] = x;
                i++;
                j--;
            }
        }
        if(left  < j)
            quickSort(item, left, j);
        if(right > i)
            quickSort(item, i, right);
    }

    static final void bubbleSort(int[] item, int left, int right) {
        int i, j, x;
        for(i = left + 1; i <= right; i++)
            for(j = right; j >= i; j--)
                if(item[j - 1] > item[j]) {
                    x  = item[j];
                    item[j] = item[j - 1];
                    item[j - 1] = x;
                }
    }

    ////////////////////////////////////////////////////////
    // Pi-computation Test

    static final String pi(int digits) {
        if (digits > 54900)
            throw new IllegalArgumentException("n must be <= 54900");

        final StringBuffer pi = new StringBuffer(digits);
        final String[] zero = { "0", "00", "000" };
        int d = 0, e, b, g, r;
        int c = (digits / 4 + 1) * 14;
        final int[] a = new int[c];
        final int f = 10000;

        Arrays.fill(a, 20000000);

        while((b = c -= 14) > 0) {
            d = e = d % f;

            while(--b > 0) {
                d = d * b + a[b];
                g = (b << 1) - 1;
                a[b] = (d % g) * f;
                d /= g;
            }

            r = e + d / f;

            if (r < 1000)
                pi.append(zero[r > 99 ? 0 : r > 9 ? 1 : 2]);
                pi.append(r);
        }
        return pi.toString();
    }

    static final long piComputationTest(int digits) throws Exception {
        long start = System.currentTimeMillis();
        String s = pi(digits);
        long stop = System.currentTimeMillis();
        return stop - start;
    }

    static final long StringTest()
    {
        long start = System.currentTimeMillis();
        StringTestWork();
        long stop = System.currentTimeMillis();
        return stop - start;
    }

    static final void StringTestWork()
    {
        final StringBuffer ss = new StringBuffer(1000);
        for(int i = 0; i < 10000000; i++)
        {
            ss.setLength(0);
            ss.append(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
            ss.append(" ");
            ss.append(" ");
            ss.append("");
            ss.append('\n');
        }
    }
    static final double FloatTestWork(int iInitVal)
    {
        double f0 = 0;
        double f1 = 123.456789;
        double f2 = 98765.12345678998765432;
        double f3 = 12345678943.98;
        for(int i = 1; i < iInitVal; i++)
        {
            f0 = (f1 / (double)i) - f2 + (f3 * (double)i);
        }
        return f0;
    }

    static final long FloatTest()
    {
       long start = System.currentTimeMillis();
       double dr = FloatTestWork(1000000000);
       long stop = System.currentTimeMillis();
       System.out.println("Result is " + dr);
       return stop - start;
    }

    static final long StrAtrTest()
    {
       long start = System.currentTimeMillis();
       StrangeAttractr(50000000);
       long stop = System.currentTimeMillis();
       return stop - start;
    }
    static final void StrangeAttractr(int iInitVal)
    {
			double x1 = -1, y1 = -1, z1 = 0;
			double x2 = -1, y2 =  1, z2 = 0;
			double x3 =  1, y3 = -1, z3 = 0;
			double x4 =  1, y4 =  1, z4 = 0;
			//  
			double px1 = 0, py1 = 0, pz1 = 0;
			double px2 = 0, py2 = 0, pz2 = 0;
			double px3 = 0, py3 = 0, pz3 = 0;
			double px4 = 0, py4 = 0, pz4 = 0;

			double ansx = 0;
			double ansy = 0;
			double ansz = 0;

			// double anssumx = 0;
			// double anssumy = 0;
			// double anssumz = 0;

			long iOptnum = iInitVal;// ,     
			double t = 0;
			double dt = (double)1 / (double)iOptnum; //   t   
			double n0 = 0 , n1 = 0, n2 = 0, n3 = 0; // . 

			for (long i = 0; i < iOptnum; i++) 
			{
		
				n0 = ( -t * (( 1 - t) * (1 - t))) / 2;
				n1 = ( 2 -5 * t * t + 3 * t * t * t) / 2;
				n2 = ( t / 2) * (1 + 4 * t - 3 * t * t);
				n3 = -(( t * t) / 2) * (1 - t);
		

				px1 = x1 * n0;
				py1 = y1	* n0;
				pz1 = z1	* n0;

				px2 = x2 * n1;
				py2 = y2	* n1;
				pz2 = z2	* n1;


				px3 = x3 * n2;
				py3 = y3	* n2;
				pz3 = z3	* n2;

				px4 = x4 * n3;
				py4 = y4	* n3;
				pz4 = z4	* n3;

				ansx = px1 + px2 + px3 + px4; 
				ansy = py1 + py2 + py3 + py4; 
				ansz = pz1 + pz2 + pz3 + pz4; 
		
				//	anssumx += ansx;
				//	anssumy += ansy;
				//	anssumz += ansz;
				t += dt;
			}

    }


}